@@ -17,6 +17,8 @@
 # Author: Jonas Borgström <jonas@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
 
+from __future__ import with_statement
+
 import doctest
 import os
 import unittest
@@ -24,14 +26,16 @@ import sys
 
 try:
     from babel import Locale
+    locale_en = Locale.parse('en_US')
 except ImportError:
-    Locale = None
+    locale_en = None
 
 from trac.config import Configuration
 from trac.core import Component, ComponentManager
 from trac.env import Environment
 from trac.db.api import _parse_db_str, DatabaseManager
 from trac.db.sqlite_backend import SQLiteConnection
+from trac.db.util import ConnectionWrapper
 import trac.db.postgres_backend
 import trac.db.mysql_backend
 from trac.ticket.default_workflow import load_workflow_config_snippet
@@ -40,9 +44,9 @@ from trac.util import translation
 
 def Mock(bases=(), *initargs, **kw):
     """
-    Simple factory for dummy classes that can be used as replacement for the 
+    Simple factory for dummy classes that can be used as replacement for the
     real implementation in tests.
-    
+
     Base classes for the mock can be specified using the first parameter, which
     must be either a tuple of class objects or a single class object. If the
     bases parameter is omitted, the base class of the mock will be object.
@@ -100,7 +104,7 @@ class MockPerm(object):
     overloading."""
 
     username = ''
-    
+
     def has_permission(self, action, realm_or_resource=None, id=False,
                        version=False):
         return True
@@ -157,81 +161,69 @@ class TestCaseSetup(unittest.TestCase):
 # -- Database utilities
 
 def get_dburi():
-    if os.environ.has_key('TRAC_TEST_DB_URI'):
-        dburi = os.environ['TRAC_TEST_DB_URI']
-        if dburi:
-            scheme, db_prop = _parse_db_str(dburi)
-            # Assume the schema 'tractest' for Postgres
-            if scheme == 'postgres' and \
-                    not db_prop.get('params', {}).get('schema'):
-                if '?' in dburi:
-                    dburi += "&schema=tractest"
-                else:
-                    dburi += "?schema=tractest"
-            return dburi
-    return 'sqlite:db/trac.db'
+    dburi = os.environ.get('TRAC_TEST_DB_URI')
+    if dburi:
+        scheme, db_prop = _parse_db_str(dburi)
+        # Assume the schema 'tractest' for Postgres
+        if scheme == 'postgres' and \
+                not db_prop.get('params', {}).get('schema'):
+            if '?' in dburi:
+                dburi += "&schema=tractest"
+            else:
+                dburi += "?schema=tractest"
+        return dburi
+    return 'sqlite::memory:'
 
 
-def reset_postgres_db(db, db_prop):
-    dbname = db.schema
-    if dbname:
-        cursor = db.cursor()
-        # reset sequences
-        # information_schema.sequences view is available in PostgreSQL 8.2+
-        # however Trac supports PostgreSQL 8.0+, uses pg_get_serial_sequence()
-        cursor.execute("""
-            SELECT sequence_name FROM (
-                SELECT pg_get_serial_sequence(%s||table_name, column_name)
-                       AS sequence_name
-                FROM information_schema.columns
-                WHERE table_schema=%s) AS tab
-            WHERE sequence_name IS NOT NULL""",
-            (dbname + '.', dbname))
-        for seq in cursor.fetchall():
-            cursor.execute('ALTER SEQUENCE %s RESTART WITH 1' % seq)
-        # clear tables
-        cursor.execute('SELECT table_name FROM information_schema.tables '
-                       'WHERE table_schema=%s', (dbname,))
-        tables = cursor.fetchall()
+def reset_sqlite_db(env, db_prop):
+    dbname = os.path.basename(db_prop['path'])
+    with env.db_transaction as db:
+        tables = db("SELECT name FROM sqlite_master WHERE type='table'")
         for table in tables:
-            # PostgreSQL supports TRUNCATE TABLE as well 
+            db("DELETE FROM %s" % table)
+        return tables
+
+
+def reset_postgres_db(env, db_prop):
+    with env.db_transaction as db:
+        dbname = db.schema
+        if dbname:
+            # reset sequences
+            # information_schema.sequences view is available in PostgreSQL 8.2+
+            # however Trac supports PostgreSQL 8.0+, uses
+            # pg_get_serial_sequence()
+            for seq in db("""
+                    SELECT sequence_name FROM (
+                        SELECT pg_get_serial_sequence(%s||table_name,
+                                                      column_name)
+                               AS sequence_name
+                        FROM information_schema.columns
+                        WHERE table_schema=%s) AS tab
+                    WHERE sequence_name IS NOT NULL""",
+                    (dbname + '.', dbname)):
+                db("ALTER SEQUENCE %s RESTART WITH 1" % seq)
+            # clear tables
+            tables = db("""SELECT table_name FROM information_schema.tables
+                           WHERE table_schema=%s""", (dbname,))
+            for table in tables:
+                db("DELETE FROM %s" % table)
+            # PostgreSQL supports TRUNCATE TABLE as well
             # (see http://www.postgresql.org/docs/8.1/static/sql-truncate.html)
             # but on the small tables used here, DELETE is actually much faster
-            cursor.execute('DELETE FROM %s' % table)
-        db.commit()
-        return tables
+            return tables
 
-def reset_mysql_db(db, db_prop):
+
+def reset_mysql_db(env, db_prop):
     dbname = os.path.basename(db_prop['path'])
     if dbname:
-        cursor = db.cursor()
-        cursor.execute('SELECT table_name FROM information_schema.tables '
-                       'WHERE table_schema=%s', (dbname,))
-        tables = cursor.fetchall()
-        for table in tables:
-            # TRUNCATE TABLE is prefered to DELETE FROM, as we need to reset
-            # the auto_increment in MySQL.
-            cursor.execute('TRUNCATE TABLE %s' % table)
-        db.commit()
-        return tables
-
-
-class InMemoryDatabase(SQLiteConnection):
-    """
-    DB-API connection object for an SQLite in-memory database, containing all
-    the default Trac tables but no data.
-    """
-    def __init__(self):
-        SQLiteConnection.__init__(self, ':memory:')
-        cursor = self.cnx.cursor()
-
-        from trac.db_default import schema
-        from trac.db.sqlite_backend import _to_sql
-        for table in schema:
-            for stmt in _to_sql(table):
-                cursor.execute(stmt)
-
-        self.cnx.commit()
+        with env.db_transaction as db:
+            tables = db("""SELECT table_name FROM information_schema.tables
+                           WHERE table_schema=%s""", (dbname,))
+            for table in tables:
+                # TRUNCATE TABLE is prefered to DELETE FROM, as we need to reset
+                # the auto_increment in MySQL.
+                db("TRUNCATE TABLE %s" % table)
+            return tables
 
 
 # -- Environment stub
@@ -240,9 +232,10 @@ class EnvironmentStub(Environment):
     """A stub of the trac.env.Environment object for testing."""
 
     href = abs_href = None
-    dbenv = db = None
+    global_databasemanager = None
 
-    def __init__(self, default_data=False, enable=None):
+    def __init__(self, default_data=False, enable=None, disable=None,
+                 path=None, destroying=False):
         """Construct a new Environment stub object.
 
         :param default_data: If True, populate the database with some
@@ -252,12 +245,15 @@ class EnvironmentStub(Environment):
         """
         ComponentManager.__init__(self)
         Component.__init__(self)
+
         self.systeminfo = []
 
         import trac
-        self.path = os.path.dirname(trac.__file__)
-        if not os.path.isabs(self.path):
-            self.path = os.path.join(os.getcwd(), self.path)
+        self.path = path
+        if self.path is None:
+            self.path = os.path.dirname(trac.__file__)
+            if not os.path.isabs(self.path):
+                self.path = os.path.join(os.getcwd(), self.path)
 
         # -- configuration
         self.config = Configuration(None)
@@ -269,21 +265,36 @@ class EnvironmentStub(Environment):
         self.config.set('logging', 'log_type', 'stderr')
         if enable is not None:
             self.config.set('components', 'trac.*', 'disabled')
+        else:
+            self.config.set('components', 'tracopt.versioncontrol.svn.*',
+                            'enabled')
         for name_or_class in enable or ():
             config_key = self._component_name(name_or_class)
             self.config.set('components', config_key, 'enabled')
+        for name_or_class in disable or ():
+            config_key = self._component_name(name_or_class)
+            self.config.set('components', config_key, 'disabled')
 
         # -- logging
         from trac.log import logger_handler_factory
         self.log, self._log_handler = logger_handler_factory('test')
 
         # -- database
+        self.config.set('components', 'trac.db.*', 'enabled')
         self.dburi = get_dburi()
-        if self.dburi.startswith('sqlite'):
-            self.config.set('trac', 'database', 'sqlite::memory:')
-            self.db = InMemoryDatabase()
 
-        if default_data:
+        init_global = False
+        if self.global_databasemanager:
+            self.components[DatabaseManager] = global_databasemanager
+        else:
+            self.config.set('trac', 'database', self.dburi)
+            self.global_databasemanager = DatabaseManager(self)
+            self.config.set('trac', 'debug_sql', True)
+            self.config.set('logging', 'log_type', 'stderr')
+            self.config.set('logging', 'log_level', 'DEBUG')
+            init_global = not destroying
+
+        if default_data or init_global:
             self.reset_db(default_data)
 
         from trac.web.href import Href
@@ -291,27 +302,7 @@ class EnvironmentStub(Environment):
         self.abs_href = Href('http://example.org/trac.cgi')
 
         self.known_users = []
-        translation.activate(Locale and Locale('en', 'US'))
-        
-    def get_read_db(self):
-        return self.get_db_cnx()
-    
-    def get_db_cnx(self, destroying=False):
-        if self.db:
-            return self.db # in-memory SQLite
-
-        # As most of the EnvironmentStubs are built at startup during
-        # the test suite formation and the creation of test cases, we can't
-        # afford to create a real db connection for each instance.
-        # So we create a special EnvironmentStub instance in charge of
-        # getting the db connections for all the other instances.
-        dbenv = EnvironmentStub.dbenv
-        if not dbenv:
-            dbenv = EnvironmentStub.dbenv = EnvironmentStub()
-            dbenv.config.set('trac', 'database', self.dburi)
-            if not destroying:
-                self.reset_db() # make sure we get rid of previous garbage
-        return DatabaseManager(dbenv).get_connection()
+        translation.activate(locale_en)
 
     def reset_db(self, default_data=None):
         """Remove all data from Trac tables, keeping the tables themselves.
@@ -319,76 +310,78 @@ class EnvironmentStub(Environment):
         :return: True upon success
         """
         from trac import db_default
-        if EnvironmentStub.dbenv:
-            db = self.get_db_cnx()
-            scheme, db_prop = _parse_db_str(self.dburi)
-
-            tables = []
-            db.rollback() # make sure there's no transaction in progress
-            try:
+        scheme, db_prop = _parse_db_str(self.dburi)
+        tables = []
+        remove_sqlite_db = False
+        try:
+            with self.db_transaction as db:
+                db.rollback() # make sure there's no transaction in progress
                 # check the database version
-                cursor = db.cursor()
-                cursor.execute("SELECT value FROM system "
-                               "WHERE name='database_version'")
-                database_version = cursor.fetchone()
+                database_version = db(
+                    "SELECT value FROM system WHERE name='database_version'")
                 if database_version:
-                    database_version = int(database_version[0])
+                    database_version = int(database_version[0][0])
                 if database_version == db_default.db_version:
                     # same version, simply clear the tables (faster)
                     m = sys.modules[__name__]
                     reset_fn = 'reset_%s_db' % scheme
                     if hasattr(m, reset_fn):
-                        tables = getattr(m, reset_fn)(db, db_prop)
+                        tables = getattr(m, reset_fn)(self, db_prop)
                 else:
                     # different version or version unknown, drop the tables
+                    remove_sqlite_db = True
                     self.destroy_db(scheme, db_prop)
-            except:
-                db.rollback()
-                # tables are likely missing
+        except Exception, e:
+            # "Database not found ...",
+            # "OperationalError: no such table: system" or the like
+            pass
 
-            if not tables:
-                del db
-                dm = DatabaseManager(EnvironmentStub.dbenv)
-                dm.init_db()
-                # we need to make sure the next get_db_cnx() will re-create 
-                # a new connection aware of the new data model - see #8518.
-                dm.shutdown() 
+        db = None # as we might shutdown the pool     FIXME no longer needed!
 
-        db = self.get_db_cnx()
-        cursor = db.cursor()
-        if default_data:
-            for table, cols, vals in db_default.get_data(db):
-                cursor.executemany("INSERT INTO %s (%s) VALUES (%s)"
+        if scheme == 'sqlite' and remove_sqlite_db:
+            path = db_prop['path']
+            if path != ':memory:':
+                if not os.path.isabs(path):
+                    path = os.path.join(self.path, path)
+                self.global_databasemanager.shutdown()
+                os.remove(path)
+
+        if not tables:
+            self.global_databasemanager.init_db()
+            # we need to make sure the next get_db_cnx() will re-create
+            # a new connection aware of the new data model - see #8518.
+            if self.dburi != 'sqlite::memory:':
+                self.global_databasemanager.shutdown()
+
+        with self.db_transaction as db:
+            if default_data:
+                for table, cols, vals in db_default.get_data(db):
+                    db.executemany("INSERT INTO %s (%s) VALUES (%s)"
                                    % (table, ','.join(cols),
                                       ','.join(['%s' for c in cols])),
                                    vals)
-        elif EnvironmentStub.dbenv:
-            cursor.execute("INSERT INTO system (name, value) "
-                           "VALUES (%s, %s)",
-                           ('database_version', str(db_default.db_version)))
-        db.commit()
+            else:
+                db("INSERT INTO system (name, value) VALUES (%s, %s)",
+                   ('database_version', str(db_default.db_version)))
 
     def destroy_db(self, scheme=None, db_prop=None):
         if not (scheme and db_prop):
             scheme, db_prop = _parse_db_str(self.dburi)
-
-        db = self.get_db_cnx(destroying=True)
-        cursor = db.cursor()
         try:
-            if scheme == 'postgres' and db.schema:
-                cursor.execute('DROP SCHEMA "%s" CASCADE' % db.schema)
-            elif scheme == 'mysql':
-                dbname = os.path.basename(db_prop['path'])
-                cursor = db.cursor()
-                cursor.execute('SELECT table_name FROM '
-                               '  information_schema.tables '
-                               'WHERE table_schema=%s', (dbname,))
-                tables = cursor.fetchall()
-                for t in tables:
-                    cursor.execute('DROP TABLE IF EXISTS `%s`' % t)
-            db.commit()
+            with self.db_transaction as db:
+                if scheme == 'postgres' and db.schema:
+                    db('DROP SCHEMA "%s" CASCADE' % db.schema)
+                elif scheme == 'mysql':
+                    dbname = os.path.basename(db_prop['path'])
+                    for table in db("""
+                          SELECT table_name FROM information_schema.tables
+                          WHERE table_schema=%s""", (dbname,)):
+                        db("DROP TABLE IF EXISTS `%s`" % table)
         except Exception:
-            db.rollback()
+            # "TracError: Database not found...",
+            # psycopg2.ProgrammingError: schema "tractest" does not exist
+            pass
+        return False
 
     # overriden
 
@@ -406,8 +399,8 @@ def locate(fn):
 
     Returns the fully-qualified path, or None.
     """
-    exec_suffix = os.name == 'nt' and '.exe' or ''
-    
+    exec_suffix = '.exe' if os.name == 'nt' else ''
+
     for p in ["."] + os.environ['PATH'].split(os.pathsep):
         f = os.path.join(p, fn + exec_suffix)
         if os.path.exists(f):
@@ -429,6 +422,9 @@ def suite():
     import trac.web.tests
     import trac.wiki.tests
     import tracopt.mimeview.tests
+    import tracopt.perm.tests
+    import tracopt.versioncontrol.git.tests
+    import tracopt.versioncontrol.svn.tests
 
     suite = unittest.TestSuite()
     suite.addTest(trac.tests.basicSuite())
@@ -444,6 +440,9 @@ def suite():
     suite.addTest(trac.web.tests.suite())
     suite.addTest(trac.wiki.tests.suite())
     suite.addTest(tracopt.mimeview.tests.suite())
+    suite.addTest(tracopt.perm.tests.suite())
+    suite.addTest(tracopt.versioncontrol.git.tests.suite())
+    suite.addTest(tracopt.versioncontrol.svn.tests.suite())
     suite.addTest(doctest.DocTestSuite(sys.modules[__name__]))
 
     return suite
