@@ -3,7 +3,7 @@
 # Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2004-2006 Christopher Lenz <cmlenz@gmx.de>
-# Copyright (C) 2005-2006 Christian Boos <cboos@neuf.fr>
+# Copyright (C) 2005-2006 Christian Boos <cboos@edgewall.org>
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -16,7 +16,9 @@
 #
 # Author: Jonas Borgström <jonas@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
-#         Christian Boos <cboos@neuf.fr>
+#         Christian Boos <cboos@edgewall.org>
+
+from __future__ import with_statement
 
 from itertools import groupby
 import os
@@ -28,13 +30,12 @@ from genshi.builder import tag
 
 from trac.config import Option, BoolOption, IntOption
 from trac.core import *
-from trac.mimeview import Context, Mimeview
+from trac.mimeview.api import Mimeview
 from trac.perm import IPermissionRequestor
 from trac.resource import Resource, ResourceNotFound
 from trac.search import ISearchSource, search_to_sql, shorten_result
 from trac.timeline.api import ITimelineEventProvider
 from trac.util import as_bool, content_disposition, embedded_numbers, pathjoin
-from trac.util.compat import any
 from trac.util.datefmt import from_utimestamp, pretty_timedelta
 from trac.util.text import exception_to_unicode, to_unicode, \
                            unicode_urlencode, shorten_line, CRLF
@@ -45,9 +46,9 @@ from trac.versioncontrol.diff import get_diff_options, diff_blocks, \
                                      unified_diff
 from trac.versioncontrol.web_ui.browser import BrowserModule
 from trac.web import IRequestHandler, RequestDone
-from trac.web.chrome import Chrome, INavigationContributor, \
-                            add_ctxtnav, add_link, add_script, \
-                            add_stylesheet, prevnext_nav
+from trac.web.chrome import (Chrome, INavigationContributor, add_ctxtnav,
+                             add_link, add_script, add_stylesheet,
+                             prevnext_nav, web_context)
 from trac.wiki import IWikiSyntaxProvider, WikiParser
 from trac.wiki.formatter import format_to
 
@@ -124,7 +125,7 @@ class ChangesetModule(Component):
                ITimelineEventProvider, IWikiSyntaxProvider, ISearchSource)
 
     property_diff_renderers = ExtensionPoint(IPropertyDiffRenderer)
-    
+
     timeline_show_files = Option('timeline', 'changeset_show_files', '0',
         """Number of files to show (`-1` for unlimited, `0` to disable).
 
@@ -143,7 +144,7 @@ class ChangesetModule(Component):
 
     timeline_collapse = BoolOption('timeline', 'changeset_collapse_events',
                                    'false',
-        """Whether consecutive changesets from the same author having 
+        """Whether consecutive changesets from the same author having
         exactly the same message should be presented as one event.
         That event will link to the range of changesets in the log view.
         (''since 0.11'')""")
@@ -160,7 +161,7 @@ class ChangesetModule(Component):
     wiki_format_messages = BoolOption('changeset', 'wiki_format_messages',
                                       'true',
         """Whether wiki formatting should be applied to changeset messages.
-        
+
         If this option is disabled, changeset messages will be rendered as
         pre-formatted text.""")
 
@@ -210,7 +211,7 @@ class ChangesetModule(Component):
         In any case, either path@rev pairs must exist.
         """
         req.perm.require('CHANGESET_VIEW')
-        
+
         # -- retrieve arguments
         full_new_path = new_path = req.args.get('new_path')
         new = req.args.get('new')
@@ -350,7 +351,7 @@ class ChangesetModule(Component):
 
         # -- HTML format
         self._render_html(req, repos, chgset, restricted, xhr, data)
-        
+
         if chgset:
             diff_params = 'new=%s' % new
         else:
@@ -371,10 +372,10 @@ class ChangesetModule(Component):
             else:
                 prevnext_nav(req, _('Previous Changeset'), _('Next Changeset'))
         else:
-            rev_href = req.href.changeset(old, full_old_path, 
+            rev_href = req.href.changeset(old, full_old_path,
                                           old=new, old_path=full_new_path)
             add_ctxtnav(req, _('Reverse Diff'), href=rev_href)
-            
+
         return 'changeset.html', data, None
 
     # Internal methods
@@ -420,8 +421,8 @@ class ChangesetModule(Component):
             title = _changeset_title(rev)
 
             # Support for revision properties (#2545)
-            context = Context.from_request(req, 'changeset', chgset.rev,
-                                           parent=repos.resource)
+            context = web_context(req, 'changeset', chgset.rev,
+                                  parent=repos.resource)
             data['context'] = context
             revprops = chgset.get_properties()
             data['properties'] = browser.render_properties('revprop', context,
@@ -438,7 +439,7 @@ class ChangesetModule(Component):
                     else:
                         prev_path = prev_rev = None
                 else:
-                    add_link(req, 'first', 
+                    add_link(req, 'first',
                              req.href.changeset(oldest_rev, reponame),
                              _('Changeset %(id)s', id=display_rev(oldest_rev)))
                     prev_path = data['old_path']
@@ -459,7 +460,7 @@ class ChangesetModule(Component):
                         else: # must be a 'D'elete or 'R'ename, show full cset
                             next_href = req.href.changeset(next_rev, reponame)
                 else:
-                    add_link(req, 'last', 
+                    add_link(req, 'last',
                              req.href.changeset(youngest_rev, reponame),
                              _('Changeset %(id)s',
                                id=display_rev(youngest_rev)))
@@ -487,7 +488,7 @@ class ChangesetModule(Component):
         def node_info(node, annotated):
             href = req.href.browser(
                 reponame, node.created_path, rev=node.created_rev,
-                annotate=annotated and 'blame' or None)
+                annotate='blame' if annotated else None)
             title = _('Show revision %(rev)s of this file in browser',
                       rev=display_rev(node.rev))
             return {'path': node.path, 'rev': node.rev,
@@ -503,8 +504,8 @@ class ChangesetModule(Component):
         def _prop_changes(old_node, new_node):
             old_props = old_node.get_properties()
             new_props = new_node.get_properties()
-            old_ctx = Context.from_request(req, old_node.resource)
-            new_ctx = Context.from_request(req, new_node.resource)
+            old_ctx = web_context(req, old_node.resource)
+            new_ctx = web_context(req, new_node.resource)
             changed_properties = []
             if old_props != new_props:
                 for k, v in sorted(old_props.items()):
@@ -580,8 +581,8 @@ class ChangesetModule(Component):
         if self.max_diff_bytes or self.max_diff_files:
             for old_node, new_node, kind, change in get_changes():
                 if change in Changeset.DIFF_CHANGES and kind == Node.FILE \
-                        and old_node.can_view(req.perm) \
-                        and new_node.can_view(req.perm):
+                        and old_node.is_viewable(req.perm) \
+                        and new_node.is_viewable(req.perm):
                     diff_files += 1
                     diff_bytes += _estimate_changes(old_node, new_node)
         show_diffs = (not self.max_diff_files or \
@@ -604,8 +605,8 @@ class ChangesetModule(Component):
         for old_node, new_node, kind, change in get_changes():
             props = []
             diffs = []
-            show_old = old_node and old_node.can_view(req.perm)
-            show_new = new_node and new_node.can_view(req.perm)
+            show_old = old_node and old_node.is_viewable(req.perm)
+            show_new = new_node and new_node.is_viewable(req.perm)
             show_entry = change != Changeset.EDIT
             show_diff = show_diffs or (new_node and new_node.path == annotated)
 
@@ -627,8 +628,8 @@ class ChangesetModule(Component):
                         'new': new_node and node_info(new_node, annotated),
                         'props': props,
                         'diffs': diffs}
-                files.append(new_node and new_node.path or \
-                             old_node and old_node.path or '')
+                files.append(new_node.path if new_node else \
+                             old_node.path if old_node else '')
                 filestats[change] += 1
                 if change in Changeset.DIFF_CHANGES:
                     if chgset:
@@ -642,7 +643,7 @@ class ChangesetModule(Component):
                             new_node.created_rev, reponame,
                             new_node.created_path,
                             old=old_node.created_rev,
-                            old_path=pathjoin(repos.reponame, 
+                            old_path=pathjoin(repos.reponame,
                                               old_node.created_path))
                         title = _('Show the %(range)s differences restricted '
                                   'to %(path)s', range='[%s:%s]' % (
@@ -659,7 +660,7 @@ class ChangesetModule(Component):
 
         data.update({'has_diffs': has_diffs, 'changes': changes, 'xhr': xhr,
                      'filestats': filestats, 'annotated': annotated,
-                     'files': files, 
+                     'files': files,
                      'location': self._get_parent_location(files),
                      'longcol': 'Revision', 'shortcol': 'r'})
 
@@ -680,7 +681,7 @@ class ChangesetModule(Component):
         req.send_response(200)
         req.send_header('Content-Type', 'text/x-patch;charset=utf-8')
         req.send_header('Content-Disposition',
-                        content_disposition('inline', filename + '.diff'))
+                        content_disposition('attachment', filename + '.diff'))
         buf = StringIO()
         mimeview = Mimeview(self.env)
 
@@ -697,7 +698,7 @@ class ChangesetModule(Component):
             new_node_info = old_node_info = ('','')
 
             if old_node:
-                if not old_node.can_view(req.perm):
+                if not old_node.is_viewable(req.perm):
                     continue
                 if mimeview.is_binary(old_node.content_type, old_node.path):
                     continue
@@ -708,7 +709,7 @@ class ChangesetModule(Component):
                 old_content = mimeview.to_unicode(old_content,
                                                   old_node.content_type)
             if new_node:
-                if not new_node.can_view(req.perm):
+                if not new_node.is_viewable(req.perm):
                     continue
                 if mimeview.is_binary(new_node.content_type, new_node.path):
                     continue
@@ -745,7 +746,7 @@ class ChangesetModule(Component):
                                          ignore_case=ignore_case,
                                          ignore_space_changes=ignore_space):
                     buf.write(line + CRLF)
-                    
+
         diff_str = buf.getvalue().encode('utf-8')
         req.send_header('Content-Length', len(diff_str))
         req.end_headers()
@@ -757,7 +758,7 @@ class ChangesetModule(Component):
         req.send_response(200)
         req.send_header('Content-Type', 'application/zip')
         req.send_header('Content-Disposition',
-                        content_disposition('inline', filename + '.zip'))
+                        content_disposition('attachment', filename + '.zip'))
 
         from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED as compression
 
@@ -767,12 +768,14 @@ class ChangesetModule(Component):
             new_path=data['new_path'], new_rev=data['new_rev'],
             old_path=data['old_path'], old_rev=data['old_rev']):
             if (kind == Node.FILE or kind == Node.DIRECTORY) and \
-                    change != Changeset.DELETE and new_node.can_view(req.perm):
+                    change != Changeset.DELETE \
+                    and new_node.is_viewable(req.perm):
                 zipinfo = ZipInfo()
                 # Note: unicode filenames are not supported by zipfile.
                 # UTF-8 is not supported by all Zip tools either,
                 # but as some do, UTF-8 is the best option here.
                 zipinfo.filename = new_node.path.strip('/').encode('utf-8')
+                zipinfo.flag_bits |= 0x800 # filename is encoded with utf-8
                 zipinfo.date_time = new_node.last_modified.utctimetuple()[:6]
                 zipinfo.compress_type = compression
                 # setting zipinfo.external_attr is needed since Python 2.5
@@ -838,12 +841,12 @@ class ChangesetModule(Component):
         if len(files) == 1:
             return files[0]
         else:
-            return '/'.join(os.path.commonprefix([f.split('/') 
+            return '/'.join(os.path.commonprefix([f.split('/')
                                                   for f in files]))
     def _get_parent_location(self, files):
         """Only get a location when there are different files,
            otherwise return the empty string."""
-        if files: 
+        if files:
             files.sort()
             prev = files[0]
             for f in files[1:]:
@@ -872,7 +875,7 @@ class ChangesetModule(Component):
                      u"\xa0\xa0-\xa0" + (repos.reponame or _('(default)')))
                     for repos in repositories
                     if not as_bool(repos.params.get('hidden'))
-                    and repos.can_view(req.perm)]
+                    and repos.is_viewable(req.perm)]
                 filters.sort()
                 add_script(req, 'common/js/timeline_multirepos.js')
                 changeset_label = _('Changesets in all repositories')
@@ -895,12 +898,12 @@ class ChangesetModule(Component):
                 show_files = int(show_files)
             else:
                 show_files = 0 # disabled
-            
+
             if self.timeline_collapse:
                 collapse_changesets = lambda c: (c.author, c.message)
             else:
                 collapse_changesets = lambda c: c.rev
-                
+
             uids_seen = {}
             def generate_changesets(repos):
                 for _, changesets in groupby(repos.get_changesets(start, stop),
@@ -909,7 +912,7 @@ class ChangesetModule(Component):
                     for cset in changesets:
                         cset_resource = Resource('changeset', cset.rev,
                                                  parent=repos.resource)
-                        if cset.can_view(req.perm):
+                        if cset.is_viewable(req.perm):
                             repos_for_uid = [repos.reponame]
                             uid = repos.get_changeset_uid(cset.rev)
                             if uid:
@@ -923,7 +926,7 @@ class ChangesetModule(Component):
                     if viewable_changesets:
                         cset = viewable_changesets[-1][0]
                         yield ('changeset', cset.date, cset.author,
-                               (viewable_changesets, 
+                               (viewable_changesets,
                                 show_location, show_files))
 
             rm = RepositoryManager(self.env)
@@ -935,7 +938,7 @@ class ChangesetModule(Component):
                             yield event
                     except TracError, e:
                         self.log.error("Timeline event provider for repository"
-                                       " '%s' failed: %r", 
+                                       " '%s' failed: %r",
                                        repos.reponame, exception_to_unicode(e))
 
     def render_timeline_event(self, context, field, event):
@@ -952,19 +955,13 @@ class ChangesetModule(Component):
             else:
                 return context.href.log(reponame or None, rev=rev_b,
                                         stop_rev=rev_a)
-            
+
         elif field == 'description':
-            branch_markup = []
-            for name, head in cset.get_branches():
-                class_ = 'branch'
-                if head:
-                    class_ += ' head'
-                branch_markup.append(tag.span(name, class_=class_))
             if self.wiki_format_messages:
                 markup = ''
                 if self.timeline_long_messages: # override default flavor
-                    context = context()
-                    context.set_hints(wiki_flavor='html', 
+                    context = context.child()
+                    context.set_hints(wiki_flavor='html',
                                       preserve_newlines=True)
             else:
                 markup = message
@@ -993,6 +990,7 @@ class ChangesetModule(Component):
                                tag.strong(self._get_location(files) or '/')),
                         markup, class_="changes")
                 elif show_files:
+                    unique_files = set()
                     for c, r, repos_for_c in changesets:
                         for chg in c.get_changes():
                             resource = c.resource.parent.child('source',
@@ -1001,15 +999,16 @@ class ChangesetModule(Component):
                                 continue
                             if show_files > 0 and len(files) > show_files:
                                 break
-                            files.append(tag.li(tag.div(class_=chg[2]),
-                                                chg[0] or '/'))
+                            unique_files.add((chg[0], chg[2]))
+                    files = [tag.li(tag.div(class_=mod), path or '/')
+                             for path, mod in sorted(unique_files)]
                     if show_files > 0 and len(files) > show_files:
                         files = files[:show_files] + [tag.li(u'\u2026')]
                     markup = tag(tag.ul(files, class_="changes"), markup)
             if message:
-                markup += format_to(self.env, None, context(cset_resource),
-                                    message)
-            return tag(branch_markup, markup)
+                markup += format_to(self.env, None,
+                                    context.child(cset_resource), message)
+            return markup
 
         single = rev_a == rev_b
         if not repos_for_uid[0]:
@@ -1017,9 +1016,9 @@ class ChangesetModule(Component):
         if reponame or len(repos_for_uid) > 1:
             title = ngettext('Changeset in %(repo)s ',
                              'Changesets in %(repo)s ',
-                             single and 1 or 2, repo=', '.join(repos_for_uid))
+                             1 if single else 2, repo=', '.join(repos_for_uid))
         else:
-            title = ngettext('Changeset ', 'Changesets ', single and 1 or 2)
+            title = ngettext('Changeset ', 'Changesets ', 1 if single else 2)
         drev_a = older_cset.repos.display_rev(rev_a)
         if single:
             title = tag(title, tag.em('[%s]' % drev_a))
@@ -1027,14 +1026,24 @@ class ChangesetModule(Component):
             drev_b = cset.repos.display_rev(rev_b)
             title = tag(title, tag.em('[%s-%s]' % (drev_a, drev_b)))
         if field == 'title':
-            return title
+            labels = []
+            for name, head in cset.get_branches():
+                if not head and name in ('default', 'master'):
+                    continue
+                class_ = 'branch'
+                if head:
+                    class_ += ' head'
+                labels.append(tag.span(name, class_=class_))
+            for name in cset.get_tags():
+                labels.append(tag.span(name, class_='tag'))
+            return title if not labels else tag(title, labels)
         elif field == 'summary':
             return _("%(title)s: %(message)s",
                      title=title, message=shorten_line(message))
-        
+
     # IWikiSyntaxProvider methods
 
-    CHANGESET_ID = r"(?:\d+|[a-fA-F\d]{8,})" # only "long enough" hexa ids
+    CHANGESET_ID = r"(?:[0-9]+|[a-fA-F0-9]{8,})" # only "long enough" hexa ids
 
     def get_wiki_syntax(self):
         yield (
@@ -1044,10 +1053,10 @@ class ChangesetModule(Component):
             # + optional query and fragment
             r"%s(?:/[^\]]*)?(?:\?[^\]]*)?(?:#[^\]]*)?\]|" % self.CHANGESET_ID +
             # r... form: allow r1 but not r1:2 (handled by the log syntax)
-            r"(?:\b|!)r\d+\b(?!:\d)(?:/[a-zA-Z0-9_/+-]+)?",
+            r"(?:\b|!)r[0-9]+\b(?!:[0-9])(?:/[a-zA-Z0-9_/+-]+)?",
             lambda x, y, z:
             self._format_changeset_link(x, 'changeset',
-                                        y[0] == 'r' and y[1:] or y[1:-1],
+                                        y[1:] if y[0] == 'r' else y[1:-1],
                                         y, z))
 
     def get_link_resolvers(self):
@@ -1081,7 +1090,7 @@ class ChangesetModule(Component):
             # rendering changeset link
             if repos:
                 changeset = repos.get_changeset(rev)
-                if changeset.can_view(formatter.perm):
+                if changeset.is_viewable(formatter.perm):
                     href = formatter.href.changeset(rev,
                                                     repos.reponame or None,
                                                     path)
@@ -1129,8 +1138,8 @@ class ChangesetModule(Component):
                                             old=data['old_rev']) + query
         return tag.a(label, class_="changeset", title=title, href=href)
 
-    # ISearchSource methods 
-    
+    # ISearchSource methods
+
     ### FIXME: move this specific implementation into cache.py
 
     def get_search_filters(self, req):
@@ -1143,24 +1152,25 @@ class ChangesetModule(Component):
         rm = RepositoryManager(self.env)
         repositories = dict((repos.params['id'], repos)
                             for repos in rm.get_real_repositories())
-        db = self.env.get_db_cnx()
-        sql, args = search_to_sql(db, ['rev', 'message', 'author'], terms)
-        cursor = db.cursor()
-        cursor.execute("SELECT repos,rev,time,author,message "
-                       "FROM revision WHERE " + sql, args)
-        for id, rev, ts, author, log in cursor:
-            try:
-                rev = int(rev)
-            except ValueError:
-                pass
-            repos = repositories.get(id)
-            if not repos:
-                continue # revisions for a no longer active repository
-            cset = repos.resource.child('changeset', rev)
-            if 'CHANGESET_VIEW' in req.perm(cset):
-                yield (req.href.changeset(rev, repos.reponame or None),
-                       '[%s]: %s' % (rev, shorten_line(log)),
-                       from_utimestamp(ts), author, shorten_result(log, terms))
+        with self.env.db_query as db:
+            sql, args = search_to_sql(db, ['rev', 'message', 'author'], terms)
+            for id, rev, ts, author, log in db("""
+                    SELECT repos, rev, time, author, message
+                    FROM revision WHERE """ + sql,
+                    args):
+                try:
+                    rev = int(rev)
+                except ValueError:
+                    pass
+                repos = repositories.get(id)
+                if not repos:
+                    continue # revisions for a no longer active repository
+                cset = repos.resource.child('changeset', rev)
+                if 'CHANGESET_VIEW' in req.perm(cset):
+                    yield (req.href.changeset(rev, repos.reponame or None),
+                           '[%s]: %s' % (rev, shorten_line(log)),
+                           from_utimestamp(ts), author,
+                           shorten_result(log, terms))
 
 
 class AnyDiffModule(Component):
@@ -1185,17 +1195,17 @@ class AnyDiffModule(Component):
 
             entries = []
             if repos:
-                entries.extend((e.isdir, e.name, 
+                entries.extend((e.isdir, e.name,
                                 '/' + pathjoin(repos.reponame, e.path))
                                for e in repos.get_node(path).get_entries()
-                               if e.can_view(req.perm))
+                               if e.is_viewable(req.perm))
             if not reponame:
                 entries.extend((True, repos.reponame, '/' + repos.reponame)
                                for repos in rm.get_real_repositories()
-                               if repos.can_view(req.perm))
+                               if repos.is_viewable(req.perm))
 
             elem = tag.ul(
-                [tag.li(isdir and tag.b(path) or path)
+                [tag.li(tag.b(path) if isdir else path)
                  for (isdir, name, path) in sorted(entries, key=kind_order)
                  if name.lower().startswith(prefix)])
 
@@ -1215,7 +1225,7 @@ class AnyDiffModule(Component):
             rm.get_repository_by_path(new_path)
         old_reponame, old_repos, old_path = \
             rm.get_repository_by_path(old_path)
-        
+
         data = {}
         if new_repos:
             data.update(new_path='/' + pathjoin(new_repos.reponame, new_path),
